<?php

   ##########################################################################
   # ZipStream - Streamed, dynamically generated zip archives.              #
   # by Paul Duncan <pabs@pablotron.org>                                    #
   #                                                                        #
   # Copyright (C) 2007-2009 Paul Duncan <pabs@pablotron.org>               #
   #                                                                        #
   # Permission is hereby granted, free of charge, to any person obtaining  #
   # a copy of this software and associated documentation files (the        #
   # "Software"), to deal in the Software without restriction, including    #
   # without limitation the rights to use, copy, modify, merge, publish,    #
   # distribute, sub-license, and/or sell copies of the Software, and to     #
   # permit persons to whom the Software is furnished to do so, subject to  #
   # the following conditions:                                              #
   #                                                                        #
   # The above copyright notice and this permission notice shall be         #
   # included in all copies or substantial portions of the of the Software. #
   #                                                                        #
   # THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,        #
   # EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF     #
   # MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. #
   # IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR      #
   # OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,  #
   # ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR  #
   # OTHER DEALINGS IN THE SOFTWARE.                                        #
   ##########################################################################

   #
   # ZipStream - Streamed, dynamically generated zip archives.
   # by Paul Duncan <pabs@pablotron.org>
   #
   # Requirements:
   #
   # * PHP version 5.1.2 or newer.
   #

   class ZipStream {
      const VERSION = '0.2.2';

      var $opt = array(),
      $files = array(),
      $cdr_ofs = 0,
      $ofs = 0,
		$size = 0;
       
      function __construct($name = null, $opt = array()) {
         # save options
         $this->opt = $opt;

         # set large file defaults: size = 20 megabytes, method = store
         if (!$this->opt['large_file_size'])
         $this->opt['large_file_size'] = 20 * 1024 * 1024;
         if (!$this->opt['large_file_method'])
         $this->opt['large_file_method'] = 'store';

         $this->output_name = $name;
         if ($name || $opt['send_http_headers'])
         $this->need_headers = true;
      }
 
      function add_file($name, $data, $opt = array()) {
         # compress data
         $zdata = gzdeflate($data);

         # calculate header attributes
         $crc  = crc32($data);
         $zlen = strlen($zdata);
         $len  = strlen($data);
         $meth = 0x08;

         # send file header
         $this->add_file_header($name, $opt, $meth, $crc, $zlen, $len);

         # print data
         $this->send($zdata);
      }
 
      function add_file_from_path($name, $path, $opt = array()) {
         if ($this->is_large_file($path)) {
            # file is too large to be read into memory; add progressively
            $this->add_large_file($name, $path, $opt);
         } else {
            # file is small enough to read into memory; read file contents and
            # handle with add_file()
            $data = file_get_contents($path);
            $this->add_file($name, $data, $opt);
         }
      }
 
      function finish() {
         # add trailing cdr record
         $this->add_cdr($this->opt);
         $this->clear();
      }

      ###################
      # PRIVATE METHODS #
      ###################

      #
      # Create and send zip header for this file.
      #
      private function add_file_header($name, $opt, $meth, $crc, $zlen, $len) {
         # strip leading slashes from file name
         # (fixes bug in windows archive viewer)
         $name = preg_replace('/^\\/+/', '', $name);

         # calculate name length
         $nlen = strlen($name);

         # create dos timestamp
         $opt['time'] = $opt['time'] ? $opt['time'] : time();
         $dts = $this->dostime($opt['time']);

         # build file header
         $fields = array(            # (from V.A of APPNOTE.TXT)
         array('V', 0x04034b50),     # local file header signature
         array('v', (6 << 8) + 3),   # version needed to extract
         array('v', 0x00),           # general purpose bit flag
         array('v', $meth),          # compresion method (deflate or store)
         array('V', $dts),           # dos timestamp
         array('V', $crc),           # crc32 of data
         array('V', $zlen),          # compressed data length
         array('V', $len),           # uncompressed data length
         array('v', $nlen),          # filename length
         array('v', 0),              # extra data len
         );

         # pack fields and calculate "total" length
         $ret = $this->pack_fields($fields);
         $cdr_len = strlen($ret) + $nlen + $zlen;

         # print header and filename
         $this->send($ret . $name);

         # add to central directory record and increment offset
         $this->add_to_cdr($name, $opt, $meth, $crc, $zlen, $len, $cdr_len);
      }

      #
      # Add a large file from the given path.
      #
      private function add_large_file($name, $path, $opt = array()) {
         $st = stat($path);
         $block_size = 1048576; # process in 1 megabyte chunks
         $algo = 'crc32b';

         # calculate header attributes
         $zlen = $len = $st['size'];

         $meth_str = $this->opt['large_file_method'];
         if ($meth_str == 'store') {
            # store method
            $meth = 0x00;
            $crc  = unpack('V', hash_file($algo, $path, true));
            $crc = $crc[1];
         } elseif ($meth_str == 'deflate') {
            # deflate method
            $meth = 0x08;

            # open file, calculate crc and compressed file length
            $fh = fopen($path, 'rb');
            $hash_ctx = hash_init($algo);
            $zlen = 0;

            # read each block, update crc and zlen
            while ($data = fgets($fh, $block_size)) {
               hash_update($hash_ctx, $data);
               $data = gzdeflate($data);
               $zlen += strlen($data);
            }

            # close file and finalize crc
            fclose($fh);
            $crc = unpack('V', hash_final($hash_ctx, true));
            $crc = $crc[1];
         } else {
            die("unknown large_file_method: $meth_str");
         }

         # send file header
         $this->add_file_header($name, $opt, $meth, $crc, $zlen, $len);

         # open input file
         $fh = fopen($path, 'rb');

         # send file blocks
         while ($data = fgets($fh, $block_size)) {
            if ($meth_str == 'deflate')
            $data = gzdeflate($data);

            # send data
            $this->send($data);
         }

         # close input file
         fclose($fh);
      }

      #
      # Is this file larger than large_file_size?
      #
      function is_large_file($path) {
         $st = stat($path);
         return ($this->opt['large_file_size'] > 0) &&
         ($st['size'] > $this->opt['large_file_size']);
      }

      #
      # Save file attributes for trailing CDR record.
      #
      private function add_to_cdr($name, $opt, $meth, $crc, $zlen, $len, $rec_len) {
         $this->files[] = array($name, $opt, $meth, $crc, $zlen, $len, $this->ofs);
         $this->ofs += $rec_len;
      }

      #
      # Send CDR record for specified file.
      #
      private function add_cdr_file($args) {
         list ($name, $opt, $meth, $crc, $zlen, $len, $ofs) = $args;

         # get attributes
         $comment = $opt['comment'] ? $opt['comment'] : '';

         # get dos timestamp
         $dts = $this->dostime($opt['time']);

         $fields = array(                  # (from V,F of APPNOTE.TXT)
         array('V', 0x02014b50),           # central file header signature
         array('v', (6 << 8) + 3),         # version made by
         array('v', (6 << 8) + 3),         # version needed to extract
         array('v', 0x00),                 # general purpose bit flag
         array('v', $meth),                # compresion method (deflate or store)
         array('V', $dts),                 # dos timestamp
         array('V', $crc),                 # crc32 of data
         array('V', $zlen),                # compressed data length
         array('V', $len),                 # uncompressed data length
         array('v', strlen($name)),        # filename length
         array('v', 0),                    # extra data len
         array('v', strlen($comment)),     # file comment length
         array('v', 0),                    # disk number start
         array('v', 0),                    # internal file attributes
         array('V', 32),                   # external file attributes
         array('V', $ofs),                 # relative offset of local header
         );

         # pack fields, then append name and comment
         $ret = $this->pack_fields($fields) . $name . $comment;

         $this->send($ret);

         # increment cdr offset
         $this->cdr_ofs += strlen($ret);
      }

      #
      # Send CDR EOF (Central Directory Record End-of-File) record.
      #
      private function add_cdr_eof($opt = null) {
         $num = count($this->files);
         $cdr_len = $this->cdr_ofs;
         $cdr_ofs = $this->ofs;

         # grab comment (if specified)
         $comment = '';
         if ($opt && $opt['comment'])
         $comment = $opt['comment'];

         $fields = array(                # (from V,F of APPNOTE.TXT)
         array('V', 0x06054b50),         # end of central file header signature
         array('v', 0x00),               # this disk number
         array('v', 0x00),               # number of disk with cdr
         array('v', $num),               # number of entries in the cdr on this disk
         array('v', $num),               # number of entries in the cdr
         array('V', $cdr_len),           # cdr size
         array('V', $cdr_ofs),           # cdr ofs
         array('v', strlen($comment)),   # zip file comment length
         );

         $ret = $this->pack_fields($fields) . $comment;
         $this->send($ret);
      }

      #
      # Add CDR (Central Directory Record) footer.
      #
      private function add_cdr($opt = null) {
         foreach ($this->files as $file)
         $this->add_cdr_file($file);
         $this->add_cdr_eof($opt);
      }

      #
      # Clear all internal variables.  Note that the stream object is not
      # usable after this.
      #
      function clear() {
         $this->files = array();
         $this->ofs = 0;
         $this->cdr_ofs = 0;
         $this->opt = array();
      }

      ###########################
      # PRIVATE UTILITY METHODS #
      ###########################

      #
      # Send HTTP headers for this stream.
      #
      private function send_http_headers() {
         # grab options
         $opt = $this->opt;

         # grab content type from options
         $content_type = 'application/x-zip';
         if ($opt['content_type'])
         $content_type = $this->opt['content_type'];

         # grab content disposition
         $disposition = 'attachment';
         if ($opt['content_disposition'])
         $disposition = $opt['content_disposition'];

         if ($this->output_name)
         $disposition .= "; filename=\"{$this->output_name}\"";

         $headers = array(
         'Content-Type'              => $content_type,
         'Content-Disposition'       => $disposition,
         'Pragma'                    => 'private',
         'Expires'                   => '0',
         'Cache-Control'             => 'maxage=1',
         'Cache-Control'             => 'private, must-revalidate, post-check=0, pre-check=0',
         'Content-Transfer-Encoding' => 'binary',
         'Accept-Encoding'				 => '0'
         );

         foreach ($headers as $key => $val)
         header("$key: $val");
      }

      #          'Content-Length'            => $this->size
      # Send string, sending HTTP headers if necessary.
      #
      private function send($str) {
         if ($this->need_headers)
         $this->send_http_headers();
         $this->need_headers = false;

         echo $str;
      }

      #
      # Convert a UNIX timestamp to a DOS timestamp.
      #
      function dostime($when = 0) {
         # get date array for timestamp
         $d = getdate($when);

         # set lower-bound on dates
         if ($d['year'] < 1980) {
            $d = array('year' => 1980, 'mon' => 1, 'mday' => 1,
            'hours' => 0, 'minutes' => 0, 'seconds' => 0);
         }

         # remove extra years from 1980
         $d['year'] -= 1980;

         # return date string
         return ($d['year'] << 25) | ($d['mon'] << 21) | ($d['mday'] << 16) |
         ($d['hours'] << 11) | ($d['minutes'] << 5) | ($d['seconds'] >> 1);
      }

      #
      # Create a format string and argument list for pack(), then call
      # pack() and return the result.
      #
      function pack_fields($fields) {
         list ($fmt, $args) = array('', array());

         # populate format string and argument list
         foreach ($fields as $field) {
            $fmt .= $field[0];
            $args[] = $field[1];
         }

         # prepend format string to argument list
         array_unshift($args, $fmt);

         # build output string from header and compressed data
         return call_user_func_array('pack', $args);
      }
   }

?>
